/**
 * (c) Copyright 2022 xsbee.
 *
 * This file is part of audcast.
 *
 * audcast is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later version.
 *
 * audcast is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this
 * program. If not, see <https://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <signal.h>
#include <errno.h>

#include <unistd.h>

#include <sys/socket.h>
#include <netdb.h>

#include <twolame.h>

#include "def.h"
#include "opts.h"
#include "util.h"

static void
transmit_audio (in, out, rate, chans, brate, vbr, vbr_level)
     int in, out, rate, chans, brate;
     bool vbr;
     float vbr_level;
{
	twolame_options * opts;

	short int * pcm_buf = NULL;
	size_t pcm_buf_cap;
	int pcm_buf_len;

	unsigned char * mp2_buf = NULL;
	int mp2_buf_cap;
	int mp2_buf_len;

	int error;

	opts = twolame_init ();

	twolame_set_in_samplerate (opts, rate);
	twolame_set_num_channels (opts, chans);

	if (vbr)
	{
		twolame_set_VBR (opts, TRUE);
		twolame_set_VBR_level (opts, vbr_level);
	}

	twolame_set_bitrate (opts, brate);

	error = twolame_init_params (opts);
	if(error != 0)
	{
		fprintf (stderr, "error: twolame_init_params(): %d\n", error);
		return;
	}

	/* https://github.com/njh/twolame/blob/master/libtwolame/twolame.c#L691
	   Therefore it is perhaps good to align to 4KiB. */
	mp2_buf_cap = round_up_int (twolame_get_framelength (opts), 4096);

	dump_mp2_config (opts, stderr);

	timer_t timer;
	struct itimerspec tspec =
	{
		.it_value =
		{
			.tv_sec = 0,
			.tv_nsec = (int) ((float) TWOLAME_SAMPLES_PER_FRAME / (float) rate * 1e+9f),
		},
		.it_interval = {0}
	};
	struct itimerspec cur_tspec;

	error = timer_create (CLOCK_REALTIME, &(struct sigevent) {
		/* No signal notification (all it has are cons). */
		.sigev_notify = SIGEV_NONE
	}, &timer);

	if (error != 0)
	{
		perror ("error: timer_create()");
		goto error;
	}

	pcm_buf_cap = TWOLAME_SAMPLES_PER_FRAME * chans * sizeof (short int);
	pcm_buf_len = 0;
	pcm_buf = malloc (pcm_buf_cap);

	if (pcm_buf == NULL)
	{
		fprintf (stderr, "error: malloc(%zu)\n", pcm_buf_cap);
		goto error;
	}

	mp2_buf_len = 0;
	mp2_buf = malloc (mp2_buf_cap);

	if (mp2_buf == NULL)
	{
		fprintf (stderr, "error: malloc(%d)\n", mp2_buf_cap);
		goto error;
	}

	for (;;)
	{
		/* Blocking FDs if return zero denote stream end. */
		pcm_buf_len = read (in, pcm_buf, pcm_buf_cap);
		if (pcm_buf_len < 0)
		{
			perror ("error: read()");
			goto error;
		}

		timer_settime (timer, 0, &tspec, NULL);

		if (pcm_buf_len == 0)
			mp2_buf_len = twolame_encode_flush (opts,
				mp2_buf,
				mp2_buf_cap);
		else
			mp2_buf_len = twolame_encode_buffer_interleaved (opts,
				pcm_buf,
				pcm_buf_len / chans / sizeof (short int),
				mp2_buf,
				mp2_buf_cap);

		if (mp2_buf_len < 0)
		{
			if (pcm_buf_len == 0)
				fprintf (stderr,
						"error: twolame_encode_flush(): %d\n", mp2_buf_len);
			else
				fprintf(stderr,
						"error: twolame_encode_buffer_interleaved(): %d\n", mp2_buf_len);
			goto error;
		}

		error = write (out, mp2_buf, mp2_buf_len);

		/* Ignore ICMP destination unreachable errors. */
		if (error < 0 && errno != ECONNREFUSED)
		{
			perror ("error: write()");
			goto error;
		}

		/* Sleep if there's time left. Blocking is not an issue. */
		timer_gettime (timer, &cur_tspec);
		nanosleep (&cur_tspec.it_value, NULL);

//		if (pcm_buf_len == 0)
//			goto error;  /* EOF */
	}

	timer_delete (timer);

error:
	if (opts != NULL)
		twolame_close(&opts);

	free (mp2_buf);
	free (pcm_buf);

	return;
}


static void
set_addrinfo_hints (struct addrinfo * hints)
{
//	hints->ai_family = AF_INET | AF_INET6;

	/* TODO: use UDP-Lite to send all *usable* DGRAMs and
			 figure out the crucial part of a MP2 frame. */
	hints->ai_socktype = SOCK_DGRAM;

	/* NOTE: multicast with IEEE 802.11 seems a bad idea, don't try. */
	hints->ai_flags = AI_PASSIVE;
}

int
main(int argc, char* argv[])
{
	struct audcast_opts opts;
	struct addrinfo hints = {0};
	struct addrinfo *res, *res0;
	int error, sock;

	parse_opts (argc, argv, &opts);
	set_addrinfo_hints (&hints);

	error = getaddrinfo (opts.hostname, opts.service, &hints, &res0);
	if (error != 0)
	{
		fprintf (stderr, "error: getaddrinfo(): %s", gai_strerror(error));
		exit (EXIT_FAILURE);
	}

	/* Attempt until valid config found. */
	for (res = res0; res; res = res->ai_next)
	{
		sock = socket (res->ai_family, res->ai_socktype, res->ai_protocol);
		if (sock < 0)
		{
			perror ("info: fail sock create attempt");
			continue;
		}

		error = connect (sock, res->ai_addr, res->ai_addrlen);
		if (error != 0)
		{
			perror ("info: fail sock conn attempt");
			continue;
		}

		break;
	}

	if (error != 0)
	{
		perror ("error: fail sock create and conn");
		exit (EXIT_FAILURE);
	}

	fprintf (stderr, "info: bound to udp://%s\n", repr_of_sockaddr(res->ai_addr));

	if (opts.broadcast)
	{
		error = setsockopt (sock, SOL_SOCKET, SO_BROADCAST, &(int) {1}, sizeof (int));
		if (error != 0)
			perror("warn: fail grant perm for brd");
	}

	freeaddrinfo (res0);

	/* Now transmit audio until reader yields. */
	transmit_audio (
		opts.input,
		sock,
		opts.samplerate,
		opts.channels,
		opts.bitrate,
		opts.vbr,
		opts.vbr_level);

	close (sock);
}
