/**
 * (c) Copyright 2022 xsbee.
 *
 * This file is part of audcast.
 *
 * audcast is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later version.
 *
 * audcast is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this
 * program. If not, see <https://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <unistd.h>
#include <fcntl.h>

#include "def.h"

#ifndef _AUDCAST_OPTS
#define _AUDCAST_OPTS

struct audcast_opts
{
	const char* hostname;
	const char* service;
	bool broadcast;

	int samplerate;
	int channels;
	int bitrate;

	bool vbr;
	float vbr_level;

	int input;  /* Input FD. */
};

static void
print_usage (char* progname)
{
	fprintf (stderr, "usage: %s"
					 " [-hBv] [-p service] [-c channels] [-r rate]"
					 " [-i input] [-q quality] [-b bitrate] [hostname]\n",
					 progname);
}

static void
print_help()
{
	fprintf (stderr, "-h	show this help message\n"
					 "-B	enable broadcast mode\n"
					 "-v	enable VBR mode (-b ignored)\n"
					 "-p	service name or port number to bind\n"
					 "-c	number of audio channels (1 or 2)\n"
					 "-r	sampling rate of audio (in hertz)\n"
					 "-i	input file to read from (default stdin)\n"
					 /* "-s	samples encoded per block (multiple of 1152)\n" */
					 /* "-S capacity of the mp2 buffer (in bytes)\n" */
					 "-q	VBR quality ([-50..50], [-10..10] effective)\n"
					 "-b	CBR bitrate (0,8,16,24,32,40,48,56,64,80,96,112,\n"
					 "                     128,144,160,192,224,256,320,384)\n");
}

static void
parse_opts (int argc, char * argv[], struct audcast_opts * opts)
{
	/* Set defaults. */
	opts->hostname = NULL;
	opts->service = AUDCAST_DEF_PORT;

	opts->input = 0;  /* stdin by default. */

	/* https://github.com/njh/twolame/blob/main/frontend/frontend.h#L31 */
	opts->samplerate = 44100;
	opts->channels = 2;
	opts->bitrate = 192;

	opts->vbr = false;
	opts->vbr_level = 5;  /* Just below -10 (VBR level). */

	opts->broadcast = false;

	const static char * optspec = "hBvp:c:r:i:q:b:";
	int ch;

	while ((ch = getopt (argc, argv, optspec)) != -1)
		switch (ch)
		{
		case 'h':
			print_usage (*argv);
			fputc ('\n', stderr);
			print_help ();
			exit (EXIT_SUCCESS);

		case 'B':
			opts->broadcast = true;
			break;

		case 'v':
			opts->vbr = true;
			break;

		case 'p':
			opts->service = optarg;
			break;

		case 'c':
			opts->channels = atoi(optarg);

			if (opts->channels < 1)
			{
				fputs ("error: channel count (<1)", stderr);
				exit (EXIT_FAILURE);
			}

			break;

		case 'r':
			opts->samplerate = atoi(optarg);

			if (opts->samplerate < 1)
			{
				fputs ("error: samplerate (<1)", stderr);
				exit (EXIT_FAILURE);
			}

			break;

		case 'i':
			opts->input = open (optarg, O_RDONLY | O_NOCTTY);

			if (opts->input < 0)
			{
				perror ("error: fail open inp file");
				exit (EXIT_FAILURE);
			}

			break;

		case 'q':
			opts->vbr_level = atof (optarg);
			break;


		case 'b':
			opts->bitrate = atoi (optarg);
			break;

		default:
			print_usage(*argv);
			exit (EXIT_FAILURE);
		}

	argc -= optind;
	argv += optind;

	/* Positional arguments. */
	if (argc >= 1)
		opts->hostname = argv[0];
}

#endif
